/* ***********************************************************
 * This file was automatically generated on 2019-09-27.      *
 *                                                           *
 * OpenHAB Bindings Version 2.1.23                           *
 *                                                           *
 * If you have a bugfix for this file and want to commit it, *
 * please fix the bug in the generator. You can find a link  *
 * to the generators git repository on tinkerforge.com       *
 *************************************************************/

package com.tinkerforge;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.Arrays;
import java.util.List;
import java.net.URI;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Collections;
import java.util.function.Function;
import java.util.function.BiConsumer;
import org.eclipse.smarthome.config.core.Configuration;
import org.eclipse.smarthome.config.core.ConfigDescription;
import org.eclipse.smarthome.config.core.ConfigDescriptionParameter.Type;
import org.eclipse.smarthome.config.core.ConfigDescriptionParameterBuilder;
import org.eclipse.smarthome.config.core.ConfigDescriptionParameterGroup;
import org.eclipse.smarthome.config.core.ParameterOption;
import org.eclipse.smarthome.core.types.State;
import org.eclipse.smarthome.core.types.StateOption;
import org.eclipse.smarthome.core.types.Command;
import org.eclipse.smarthome.core.types.CommandDescriptionBuilder;
import org.eclipse.smarthome.core.types.CommandOption;
import org.eclipse.smarthome.core.thing.ThingTypeUID;
import org.eclipse.smarthome.core.thing.type.ChannelDefinitionBuilder;
import org.eclipse.smarthome.core.thing.type.ChannelType;
import org.eclipse.smarthome.core.thing.type.ChannelTypeBuilder;
import org.eclipse.smarthome.core.thing.type.ChannelTypeUID;
import org.eclipse.smarthome.core.thing.type.ThingType;
import org.eclipse.smarthome.core.thing.type.ThingTypeBuilder;
import org.eclipse.smarthome.core.types.StateDescriptionFragmentBuilder;
import org.eclipse.smarthome.binding.tinkerforge.internal.TinkerforgeBindingConstants;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.eclipse.smarthome.core.thing.CommonTriggerEvents;
import org.eclipse.smarthome.core.library.types.OnOffType;
import org.eclipse.smarthome.core.library.types.StringType;

/**
 * Two relays to switch AC/DC devices
 */
public class BrickletDualRelay extends Device {
	public final static int DEVICE_IDENTIFIER = 26;
	public final static String DEVICE_DISPLAY_NAME = "Dual Relay Bricklet";

    public final static DeviceInfo DEVICE_INFO = new DeviceInfo(DEVICE_DISPLAY_NAME, "dualrelay", DEVICE_IDENTIFIER, BrickletDualRelay.class);

	public final static byte FUNCTION_SET_STATE = (byte)1;
	public final static byte FUNCTION_GET_STATE = (byte)2;
	public final static byte FUNCTION_SET_MONOFLOP = (byte)3;
	public final static byte FUNCTION_GET_MONOFLOP = (byte)4;
	public final static byte FUNCTION_SET_SELECTED_STATE = (byte)6;
	public final static byte FUNCTION_GET_IDENTITY = (byte)255;
	private final static int CALLBACK_MONOFLOP_DONE = 5;


	private List<MonoflopDoneListener> listenerMonoflopDone = new CopyOnWriteArrayList<MonoflopDoneListener>();

	public class State {
		public boolean relay1;
		public boolean relay2;

		public String toString() {
			return "[" + "relay1 = " + relay1 + ", " + "relay2 = " + relay2 + "]";
		}
	}

	public class Monoflop {
		public boolean state;
		public long time;
		public long timeRemaining;

		public String toString() {
			return "[" + "state = " + state + ", " + "time = " + time + ", " + "timeRemaining = " + timeRemaining + "]";
		}
	}

	/**
	 * This listener is triggered whenever a monoflop timer reaches 0. The
	 * parameter contain the relay (1 or 2) and the current state of the relay
	 * (the state after the monoflop).
	 */
	public interface MonoflopDoneListener extends DeviceListener {
		public void monoflopDone(short relay, boolean state);
	}

	/**
	 * Creates an object with the unique device ID \c uid. and adds it to
	 * the IP Connection \c ipcon.
	 */
	public BrickletDualRelay(String uid, IPConnection ipcon) {
		super(uid, ipcon);

		apiVersion[0] = 2;
		apiVersion[1] = 0;
		apiVersion[2] = 0;
		responseExpected[IPConnection.unsignedByte(FUNCTION_SET_STATE)] = RESPONSE_EXPECTED_FLAG_FALSE;
		responseExpected[IPConnection.unsignedByte(FUNCTION_GET_STATE)] = RESPONSE_EXPECTED_FLAG_ALWAYS_TRUE;
		responseExpected[IPConnection.unsignedByte(FUNCTION_SET_MONOFLOP)] = RESPONSE_EXPECTED_FLAG_FALSE;
		responseExpected[IPConnection.unsignedByte(FUNCTION_GET_MONOFLOP)] = RESPONSE_EXPECTED_FLAG_ALWAYS_TRUE;
		responseExpected[IPConnection.unsignedByte(FUNCTION_SET_SELECTED_STATE)] = RESPONSE_EXPECTED_FLAG_FALSE;
		responseExpected[IPConnection.unsignedByte(FUNCTION_GET_IDENTITY)] = RESPONSE_EXPECTED_FLAG_ALWAYS_TRUE;

		callbacks[CALLBACK_MONOFLOP_DONE] = new IPConnection.DeviceCallbackListener() {
			public void callback(byte[] packet) {
				ByteBuffer bb = ByteBuffer.wrap(packet, 8, packet.length - 8);
				bb.order(ByteOrder.LITTLE_ENDIAN);

				short relay = IPConnection.unsignedByte(bb.get());
				boolean state = (bb.get()) != 0;

				for (MonoflopDoneListener listener: listenerMonoflopDone) {
					listener.monoflopDone(relay, state);
				}
			}
		};
	}

	/**
	 * Sets the state of the relays, *true* means on and *false* means off.
	 * For example: (true, false) turns relay 1 on and relay 2 off.
	 * 
	 * If you just want to set one of the relays and don't know the current state
	 * of the other relay, you can get the state with {@link BrickletDualRelay#getState()} or you
	 * can use {@link BrickletDualRelay#setSelectedState(short, boolean)}.
	 * 
	 * All running monoflop timers will be aborted if this function is called.
	 * 
	 * The default value is (*false*, *false*).
	 */
	public void setState(boolean relay1, boolean relay2) throws TinkerforgeException {
		ByteBuffer bb = ipcon.createRequestPacket((byte)10, FUNCTION_SET_STATE, this);

		bb.put((byte)(relay1 ? 1 : 0));
		bb.put((byte)(relay2 ? 1 : 0));

		sendRequest(bb.array());
	}

	/**
	 * Returns the state of the relays, *true* means on and *false* means off.
	 */
	public State getState() throws TinkerforgeException {
		ByteBuffer bb = ipcon.createRequestPacket((byte)8, FUNCTION_GET_STATE, this);


		byte[] response = sendRequest(bb.array());

		bb = ByteBuffer.wrap(response, 8, response.length - 8);
		bb.order(ByteOrder.LITTLE_ENDIAN);

		State obj = new State();
		obj.relay1 = (bb.get()) != 0;
		obj.relay2 = (bb.get()) != 0;

		return obj;
	}

	/**
	 * The first parameter can be 1 or 2 (relay 1 or relay 2). The second parameter
	 * is the desired state of the relay (*true* means on and *false* means off).
	 * The third parameter indicates the time (in ms) that the relay should hold
	 * the state.
	 * 
	 * If this function is called with the parameters (1, true, 1500):
	 * Relay 1 will turn on and in 1.5s it will turn off again.
	 * 
	 * A monoflop can be used as a failsafe mechanism. For example: Lets assume you
	 * have a RS485 bus and a Dual Relay Bricklet connected to one of the slave
	 * stacks. You can now call this function every second, with a time parameter
	 * of two seconds. The relay will be on all the time. If now the RS485
	 * connection is lost, the relay will turn off in at most two seconds.
	 */
	public void setMonoflop(short relay, boolean state, long time) throws TinkerforgeException {
		ByteBuffer bb = ipcon.createRequestPacket((byte)14, FUNCTION_SET_MONOFLOP, this);

		bb.put((byte)relay);
		bb.put((byte)(state ? 1 : 0));
		bb.putInt((int)time);

		sendRequest(bb.array());
	}

	/**
	 * Returns (for the given relay) the current state and the time as set by
	 * {@link BrickletDualRelay#setMonoflop(short, boolean, long)} as well as the remaining time until the state flips.
	 * 
	 * If the timer is not running currently, the remaining time will be returned
	 * as 0.
	 */
	public Monoflop getMonoflop(short relay) throws TinkerforgeException {
		ByteBuffer bb = ipcon.createRequestPacket((byte)9, FUNCTION_GET_MONOFLOP, this);

		bb.put((byte)relay);

		byte[] response = sendRequest(bb.array());

		bb = ByteBuffer.wrap(response, 8, response.length - 8);
		bb.order(ByteOrder.LITTLE_ENDIAN);

		Monoflop obj = new Monoflop();
		obj.state = (bb.get()) != 0;
		obj.time = IPConnection.unsignedInt(bb.getInt());
		obj.timeRemaining = IPConnection.unsignedInt(bb.getInt());

		return obj;
	}

	/**
	 * Sets the state of the selected relay (1 or 2), *true* means on and *false* means off.
	 * 
	 * A running monoflop timer for the selected relay will be aborted if this function is called.
	 * 
	 * The other relay remains untouched.
	 */
	public void setSelectedState(short relay, boolean state) throws TinkerforgeException {
		ByteBuffer bb = ipcon.createRequestPacket((byte)10, FUNCTION_SET_SELECTED_STATE, this);

		bb.put((byte)relay);
		bb.put((byte)(state ? 1 : 0));

		sendRequest(bb.array());
	}

	/**
	 * Returns the UID, the UID where the Bricklet is connected to,
	 * the position, the hardware and firmware version as well as the
	 * device identifier.
	 * 
	 * The position can be 'a', 'b', 'c' or 'd'.
	 * 
	 * The device identifier numbers can be found :ref:`here &lt;device_identifier&gt;`.
	 * |device_identifier_constant|
	 */
	public Identity getIdentity() throws TinkerforgeException {
		ByteBuffer bb = ipcon.createRequestPacket((byte)8, FUNCTION_GET_IDENTITY, this);


		byte[] response = sendRequest(bb.array());

		bb = ByteBuffer.wrap(response, 8, response.length - 8);
		bb.order(ByteOrder.LITTLE_ENDIAN);

		Identity obj = new Identity();
		obj.uid = IPConnection.string(bb, 8);
		obj.connectedUid = IPConnection.string(bb, 8);
		obj.position = (char)(bb.get());
		for (int i = 0; i < 3; i++) {
			obj.hardwareVersion[i] = IPConnection.unsignedByte(bb.get());
		}
		for (int i = 0; i < 3; i++) {
			obj.firmwareVersion[i] = IPConnection.unsignedByte(bb.get());
		}
		obj.deviceIdentifier = IPConnection.unsignedShort(bb.getShort());

		return obj;
	}

	/**
	 * Adds a MonoflopDone listener.
	 */
	public void addMonoflopDoneListener(MonoflopDoneListener listener) {
		listenerMonoflopDone.add(listener);
	}

	/**
	 * Removes a MonoflopDone listener.
	 */
	public void removeMonoflopDoneListener(MonoflopDoneListener listener) {
		listenerMonoflopDone.remove(listener);
	}

    private final Logger logger = LoggerFactory.getLogger(BrickletDualRelay.class);
    private final static Logger static_logger = LoggerFactory.getLogger(BrickletDualRelay.class);

    @Override
    public void initialize(org.eclipse.smarthome.config.core.Configuration config, Function<String, org.eclipse.smarthome.config.core.Configuration> getChannelConfigFn, BiConsumer<String, org.eclipse.smarthome.core.types.State> updateStateFn, BiConsumer<String, String> triggerChannelFn) throws TinkerforgeException {
        BrickletDualRelayConfig cfg = (BrickletDualRelayConfig) config.as(BrickletDualRelayConfig.class);
        this.addMonoflopDoneListener((relay, state) -> {if(relay == 1) {updateStateFn.accept("DualRelayRelay1", transformDualRelayRelay1Callback0(relay, state, cfg));}});
		this.addMonoflopDoneListener((relay, state) -> {if(relay == 2) {updateStateFn.accept("DualRelayRelay2", transformDualRelayRelay2Callback0(relay, state, cfg));}});
        
		{
		BrickletDualRelayRelayConfig channelCfg = getChannelConfigFn.apply("DualRelayRelay1").as(BrickletDualRelayRelayConfig.class);
		
		}
		{
		BrickletDualRelayRelayConfig channelCfg = getChannelConfigFn.apply("DualRelayRelay2").as(BrickletDualRelayRelayConfig.class);
		
		}
		{
		BrickletDualRelayMonoflopConfig channelCfg = getChannelConfigFn.apply("DualRelayMonofloprelay1").as(BrickletDualRelayMonoflopConfig.class);
		
		}
		{
		BrickletDualRelayMonoflopConfig channelCfg = getChannelConfigFn.apply("DualRelayMonofloprelay2").as(BrickletDualRelayMonoflopConfig.class);
		
		}
    }

    @Override
    public void dispose(org.eclipse.smarthome.config.core.Configuration config) throws TinkerforgeException {
        BrickletDualRelayConfig cfg = (BrickletDualRelayConfig) config.as(BrickletDualRelayConfig.class);
        this.listenerMonoflopDone.clear();
		this.listenerMonoflopDone.clear();
        
		
		
    }

    @Override
    public List<String> getEnabledChannels(org.eclipse.smarthome.config.core.Configuration config) throws TinkerforgeException{
        BrickletDualRelayConfig cfg = (BrickletDualRelayConfig) config.as(BrickletDualRelayConfig.class);
        List<String> result = new ArrayList<String>();
        result.add("DualRelayRelay1");
		result.add("DualRelayRelay2");
		result.add("DualRelayMonofloprelay1");
		result.add("DualRelayMonofloprelay2");
        return result;
    }

    public static ChannelType getChannelType(ChannelTypeUID channelTypeUID) {
        switch(channelTypeUID.getId()) {
            case "DualRelayRelay":
                return ChannelTypeBuilder.state(new ChannelTypeUID("tinkerforge", "DualRelayRelay"), "NOT USED", "Switch").withConfigDescriptionURI(URI.create("channel-type:tinkerforge:DualRelayRelay")).withDescription("NOT USED").build();
            case "DualRelayMonoflop":
                return ChannelTypeBuilder.state(new ChannelTypeUID("tinkerforge", "DualRelayMonoflop"), "NOT USED", "String").withConfigDescriptionURI(URI.create("channel-type:tinkerforge:DualRelayMonoflop")).withDescription("Triggers a monoflop as configured")
.withCommandDescription(CommandDescriptionBuilder.create().withCommandOption(new CommandOption("TRIGGER", "Trigger")).build()).build();
            default:
                static_logger.debug("Unknown channel type ID {}", channelTypeUID.getId());
                break;
        }

        return null;
    }

    public static ThingType getThingType(ThingTypeUID thingTypeUID) {
        return ThingTypeBuilder.instance(thingTypeUID, "Tinkerforge Dual Relay Bricklet").isListed(false).withSupportedBridgeTypeUIDs(Arrays.asList(TinkerforgeBindingConstants.THING_TYPE_BRICK_DAEMON.getId())).withConfigDescriptionURI(URI.create("thing-type:tinkerforge:" + thingTypeUID.getId())).withDescription("Two relays to switch AC/DC devices").withChannelDefinitions(Arrays.asList(new ChannelDefinitionBuilder("DualRelayRelay1", new ChannelTypeUID("tinkerforge", "DualRelayRelay")).withLabel("Relay 1").withDescription("Switches Relay 1. A running monoflop timer for this relay will be aborted if the relay is toggled by this channel.").build(), new ChannelDefinitionBuilder("DualRelayRelay2", new ChannelTypeUID("tinkerforge", "DualRelayRelay")).withLabel("Relay 2").withDescription("Switches Relay 2. A running monoflop timer for this relay will be aborted if the relay is toggled by this channel.").build(), new ChannelDefinitionBuilder("DualRelayMonofloprelay1", new ChannelTypeUID("tinkerforge", "DualRelayMonoflop")).withLabel("Monoflop Relay 1").build(), new ChannelDefinitionBuilder("DualRelayMonofloprelay2", new ChannelTypeUID("tinkerforge", "DualRelayMonoflop")).withLabel("Monoflop Relay 2").build())).build();
    }

    public static ConfigDescription getConfigDescription(URI uri) {
        switch(uri.toASCIIString()) {
            case "thing-type:tinkerforge:dualrelay":
                return new ConfigDescription(uri, Arrays.asList(), Arrays.asList(new ConfigDescriptionParameterGroup("update_intervals", "null", true, "Update Intervals", "Update Intervals")));
            case "channel-type:tinkerforge:DualRelayRelay":
                return new ConfigDescription(uri, Arrays.asList(ConfigDescriptionParameterBuilder.create("updateInterval", Type.INTEGER).withDefault("1000").withDescription("Specifies the update interval in milliseconds. A value of 0 disables automatic updates.").withGroupName("update_intervals").withLabel("Update Interval").withUnit("ms").build()));
            case "channel-type:tinkerforge:DualRelayMonoflop":
                return new ConfigDescription(uri, Arrays.asList(ConfigDescriptionParameterBuilder.create("monoflopDuration", Type.INTEGER).withDefault("1000").withDescription("The time (in ms) that the relay should hold the configured value.").withLabel("Monoflop Duration").withUnit("ms").withMinimum(BigDecimal.valueOf(0)).withMaximum(BigDecimal.valueOf(2147483647)).build(), ConfigDescriptionParameterBuilder.create("monoflopValue", Type.BOOLEAN).withDefault("true").withDescription("The desired value of the specified channel. Activated means relay closed and Deactivated means relay open.").withLabel("Monoflop Value").build()));
            default:
                static_logger.debug("Unknown config description URI {}", uri.toASCIIString());
                break;
        }
        return null;
    }

        @Override
    public void refreshValue(String value, org.eclipse.smarthome.config.core.Configuration config, org.eclipse.smarthome.config.core.Configuration channelConfig, BiConsumer<String, org.eclipse.smarthome.core.types.State> updateStateFn, BiConsumer<String, String> triggerChannelFn) throws TinkerforgeException {
        BrickletDualRelayConfig cfg = (BrickletDualRelayConfig) config.as(BrickletDualRelayConfig.class);
        switch(value) {
            case "DualRelayRelay1": {
                   BrickletDualRelayRelayConfig channelCfg = channelConfig.as(BrickletDualRelayRelayConfig.class);
                   updateStateFn.accept(value, transformDualRelayRelay1Getter0(this.getState(), cfg));
                   return;
               }
            case "DualRelayRelay2": {
                   BrickletDualRelayRelayConfig channelCfg = channelConfig.as(BrickletDualRelayRelayConfig.class);
                   updateStateFn.accept(value, transformDualRelayRelay2Getter0(this.getState(), cfg));
                   return;
               }
            case "DualRelayMonofloprelay1": {
                   BrickletDualRelayMonoflopConfig channelCfg = channelConfig.as(BrickletDualRelayMonoflopConfig.class);
                   updateStateFn.accept(value, transformDualRelayMonofloprelay1Getter0(this.getMonoflop((short) 1), cfg));
                   return;
               }
            case "DualRelayMonofloprelay2": {
                   BrickletDualRelayMonoflopConfig channelCfg = channelConfig.as(BrickletDualRelayMonoflopConfig.class);
                   updateStateFn.accept(value, transformDualRelayMonofloprelay2Getter0(this.getMonoflop((short) 2), cfg));
                   return;
               }
            default:
                logger.warn("Refresh for unknown channel {}", value);
                break;
        }
    }
    

        @Override
    public List<SetterRefresh> handleCommand(org.eclipse.smarthome.config.core.Configuration config, org.eclipse.smarthome.config.core.Configuration channelConfig, String channel, Command command) throws TinkerforgeException {
        List<SetterRefresh> result = new ArrayList<SetterRefresh>();
        BrickletDualRelayConfig cfg = (BrickletDualRelayConfig) config.as(BrickletDualRelayConfig.class);
        switch(channel) {
            case "DualRelayRelay1":
                if (command instanceof OnOffType) {
                    BrickletDualRelayRelayConfig channelCfg = channelConfig.as(BrickletDualRelayRelayConfig.class);
                    OnOffType cmd = (OnOffType) command;
                    this.setSelectedState((short)1, cmd == OnOffType.ON);
                    
                } else {
                    logger.warn("Command type {} not supported for channel {}. Please use a OnOffType.", command.getClass().getName(), channel);
                }
                break;
            case "DualRelayRelay2":
                if (command instanceof OnOffType) {
                    BrickletDualRelayRelayConfig channelCfg = channelConfig.as(BrickletDualRelayRelayConfig.class);
                    OnOffType cmd = (OnOffType) command;
                    this.setSelectedState((short)2, cmd == OnOffType.ON);
                    
                } else {
                    logger.warn("Command type {} not supported for channel {}. Please use a OnOffType.", command.getClass().getName(), channel);
                }
                break;
            case "DualRelayMonofloprelay1":
                if (command instanceof StringType) {
                    BrickletDualRelayMonoflopConfig channelCfg = channelConfig.as(BrickletDualRelayMonoflopConfig.class);
                    StringType cmd = (StringType) command;
                    this.setMonoflop((short)1, channelCfg.monoflopValue.booleanValue(), channelCfg.monoflopDuration);
                    result.add(new SetterRefresh("DualRelayRelay1", 0));
                } else {
                    logger.warn("Command type {} not supported for channel {}. Please use a StringType.", command.getClass().getName(), channel);
                }
                break;
            case "DualRelayMonofloprelay2":
                if (command instanceof StringType) {
                    BrickletDualRelayMonoflopConfig channelCfg = channelConfig.as(BrickletDualRelayMonoflopConfig.class);
                    StringType cmd = (StringType) command;
                    this.setMonoflop((short)2, channelCfg.monoflopValue.booleanValue(), channelCfg.monoflopDuration);
                    result.add(new SetterRefresh("DualRelayRelay2", 0));
                } else {
                    logger.warn("Command type {} not supported for channel {}. Please use a StringType.", command.getClass().getName(), channel);
                }
                break;
            default:
                logger.warn("Command for unknown channel {}", channel);
        }
        return result;
    }

        private org.eclipse.smarthome.core.types.State transformDualRelayRelay1Callback0(short relay, boolean state, BrickletDualRelayConfig cfg) {
        return state ? OnOffType.ON : OnOffType.OFF;
    }
	    private org.eclipse.smarthome.core.types.State transformDualRelayRelay2Callback0(short relay, boolean state, BrickletDualRelayConfig cfg) {
        return state ? OnOffType.ON : OnOffType.OFF;
    }
	    private org.eclipse.smarthome.core.types.State transformDualRelayRelay1Getter0(State value, BrickletDualRelayConfig cfg) {
        return value.relay1 ? OnOffType.ON : OnOffType.OFF;
    }
	    private org.eclipse.smarthome.core.types.State transformDualRelayRelay2Getter0(State value, BrickletDualRelayConfig cfg) {
        return value.relay2 ? OnOffType.ON : OnOffType.OFF;
    }
	    private org.eclipse.smarthome.core.types.State transformDualRelayMonofloprelay1Getter0(Monoflop value, BrickletDualRelayConfig cfg) {
        return value.state ? OnOffType.ON : OnOffType.OFF;
    }
	    private org.eclipse.smarthome.core.types.State transformDualRelayMonofloprelay2Getter0(Monoflop value, BrickletDualRelayConfig cfg) {
        return value.state ? OnOffType.ON : OnOffType.OFF;
    }
}
